import {createWorker, PSM, ImageLike, Page, Word, Bbox, Line, Paragraph} from 'tesseract.js';
import throttle from 'lodash.throttle';

export enum LoadingStatus {
    CORE_LOADING = 'loading tesseract core',
    OCR_INITIALIZATION = 'initializing tesseract',
    LANGUAGE_LOADING = 'loading language traineddata',
    API_INITIALIZATION = 'initializing api',
    RECOGNIZING = 'recognizing text'
}

type ProgressListener = (event: { status: LoadingStatus, progress: number }) => void;
type Options = {
    language?: string,
    lineClassName?: string,
    wordClassName?: string
}

const THROTTLE_DELAY = 100;
const WORD_CLASS_NAME = 'text-recognition-word';

export default class Ocr {
    data: Page = null;
    language: string = null;
    lineClassName: string = '';
    wordClassName: string = '';
    progressListeners: ProgressListener[] = []
    minFontSize = 2;

    /**
     * In the constructor you can define some default options.
     *
     * @param {Options} options Options for general configuration.
     */
    constructor({language = 'eng', lineClassName = '', wordClassName = ''}: Options = {}) {
        this.language = language;
        this.lineClassName = lineClassName;
        this.wordClassName = wordClassName;
    }

    /**
     * Add event listeners for highlighting selected blocks and format selected text.
     */
    addSelectionListener() {
        document.addEventListener('selectionchange', throttle(this.handleSelection, THROTTLE_DELAY));
        document.addEventListener('mousedown', this.handleMouseDown);
        document.addEventListener('copy', this.handleCopy);
    }

    /**
     * Removing event listeners for highlighting selected blocks and format selected text.
     */
    removeSelectionListener() {
        document.removeEventListener('selectionchange', throttle(this.handleSelection, THROTTLE_DELAY));
        document.removeEventListener('mousedown', this.handleMouseDown);
        document.removeEventListener('copy', this.handleCopy);
    }

    /**
     * Adding progress listener for ocr.
     *
     * @param {ProgressListener} listener Callback function.
     */
    addProgressListener(listener: ProgressListener) {
        this.progressListeners.push(listener);
    }

    /**
     * Removing progress listener for ocr.
     *
     * @param {ProgressListener} listener Callback function that was added before.
     */
    removeProgressListener(listener: ProgressListener) {
        this.progressListeners = this.progressListeners.filter(item => item !== listener);
    }

    /**
     * Parse image with ocr.
     *
     * @param {ImageLike} image The image instance.
     */
    async parseImage(image: ImageLike) {
        const worker = createWorker({
            logger: m => {
                this.progressListeners.forEach(callback => {
                    callback(m)
                })
            },
        });
        await worker.load();
        await worker.loadLanguage(this.language);
        await worker.initialize(this.language);
        await worker.setParameters({
            tessedit_pageseg_mode: PSM.AUTO_ONLY,
        });
        const {data} = await worker.recognize(image);
        this.data = data;
        await worker.terminate();
    }

    private updateTextSize(el: HTMLElement) {
        const wrap = document.createElement('span');
        document.body.appendChild(wrap);

        wrap.innerHTML = el.textContent.trim();
        const height = parseInt(el.style.height.replace('px', ''));
        const width = parseInt(el.style.width.replace('px', ''));
        let fontSize = height;
        if (fontSize > this.minFontSize) {
            wrap.style.fontSize = `${fontSize}px`;
            const fontSizeH = fontSize * height / wrap.offsetHeight;
            const fontSizeW = fontSize * width / wrap.offsetWidth;
            fontSize = fontSizeH < fontSizeW ? fontSizeH : fontSizeW;
        }
        if (fontSize < this.minFontSize) {
            fontSize = this.minFontSize;
        }
        el.style.fontSize = `${fontSize}px`;

        wrap.remove();
    }

    private updateElementSize(el: HTMLElement, {x0, x1, y0, y1}: Bbox, scale: number) {
        el.style.width = `${(x1 - x0) * scale}px`;
        el.style.height = `${(y1 - y0) * scale}px`;
    }

    private mapWords(words: Word[], scale: number, parentBbox: Bbox) {
        const wordElementsList = [];

        for (let i = 0; i < words.length; i++) {
            const {bbox, text} = words[i];
            const nextBboxX = i < words.length - 1 ? words[i + 1].bbox.x0 : bbox.x1;
            const elementBbox = {x0: bbox.x0, x1: nextBboxX, y0: parentBbox.y0, y1: parentBbox.y1}

            if (!text.trim().length) continue;

            const el = document.createElement('span');
            this.updateElementSize(el, elementBbox, scale);
            el.innerText = text;
            el.classList.add(WORD_CLASS_NAME);
            if (this.wordClassName.length) {
                el.classList.add(this.wordClassName);
            }

            this.updateTextSize(el)

            wordElementsList.push(el);
        }

        return wordElementsList;
    }

    private mapLine(line: Line, scale: number) {
        const {bbox, words} = line;

        const el = document.createElement('span');
        this.updateElementSize(el, bbox, scale);
        el.style.display = 'flex';

        if (this.lineClassName.length) {
            el.classList.add(this.lineClassName);
        }

        const wordElements = this.mapWords(words, scale, bbox);

        if (!wordElements.length) return null;

        wordElements.forEach(word => {
            el.appendChild(word);
        })

        return {element: el, bbox};
    }

    private mapParagraph(paragraph: Paragraph, scale: number) {
        const {bbox, lines} = paragraph;

        const el = document.createElement('p');
        this.updateElementSize(el, bbox, scale)
        el.style.position = 'absolute';
        el.style.left = `${bbox.x0 * scale}px`;
        el.style.top = `${bbox.y0 * scale}px`;
        el.style.margin = '0px';


        const lineElements: { element: HTMLSpanElement, bbox: Bbox }[] = [];
        for (let i = 0; i < lines.length; i++) {
            const line = this.mapLine(lines[i], scale);
            if (line) {
                lineElements.push(line);
            }
        }

        if (!lineElements.length) return null;

        for (let i = 0; i < lineElements.length; i++) {
            const {element, bbox} = lineElements[i];
            if (i > 0) {
                const marginTop = bbox.y0 - lineElements[i - 1].bbox.y1;
                element.style.marginTop = `${marginTop}px`;
            }
            el.appendChild(element);
        }

        return {element: el, bbox};
    }

    /**
     * Append generated html to the container.
     *
     * @param {HTMLElement} container The container where elements should append.
     * @param {number} scale Scale of the container compared to original image size.
     */
    appendHtml(container: HTMLElement, scale = 1) {
        if (!this.data.hocr) return;

        let elements: { bbox: Bbox, element: HTMLParagraphElement }[] = []
        const {paragraphs} = this.data;
        for (let i = 0; i < paragraphs.length; i++) {
            const paragraph = this.mapParagraph(paragraphs[i], scale)
            if (paragraph) {
                elements.push(paragraph);
            }
        }

        for (let i = 0; i < elements.length; i++) {
            const {element, bbox} = elements[i];

            if (i < elements.length - 1) {
                const padding = elements[i + 1].bbox.y0 - bbox.y1;
                if (padding > 0) {
                    element.style.paddingBottom = `${padding}px`;
                }
            }

            container.appendChild(element)
        }
    }

    private handleSelection() {
        const selection = document.getSelection();
        if (selection?.focusNode?.parentNode) {
            const selectableElements = document.querySelectorAll(`.${WORD_CLASS_NAME}`);
            for (const element of Array.from(selectableElements)) {
                if (!!selection.toString() && selection.containsNode(element, true)) {
                    (element as HTMLElement).style.background = '#ACCEF7';
                } else {
                    (element as HTMLElement).style.background = 'none';
                }
            }
        }
    }

    private handleMouseDown() {
        const selection = document.getSelection();
        if (!selection) return;
        const selectableElements = document.querySelectorAll(`.${WORD_CLASS_NAME}`);
        for (const element of Array.from(selectableElements)) {
            if (selection.containsNode(element, true)) {
                (element as HTMLElement).style.background = 'none';
            }
        }
    }

    private handleCopy(event) {
        const selection = document.getSelection();
        if (!selection) return;
        const selectableElements = document.querySelectorAll(`.${WORD_CLASS_NAME}`);
        let output = '';
        for (const element of Array.from(selectableElements)) {
            if (selection.containsNode(element, true)) {
                output += `${(element as HTMLElement).textContent} `;
                const blockChildNodes = element.closest('p').childNodes;
                const lineChildNodes = blockChildNodes[blockChildNodes.length - 1].childNodes;
                const lastLineChildNode = lineChildNodes[lineChildNodes.length - 1];
                if (lastLineChildNode.isEqualNode(element)) {
                    output += `\n`;
                }
            }
        }
        event.clipboardData.setData('text/plain', output);
        event.preventDefault();
    }
}
